/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// Main windowed and fullscreen graphics interface module. This module
// is used for both the software and OpenGL rendering versions of the
// Quake refresh engine.

#include <dlfcn.h> // ELF dl loader
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include "../client/cl_local.h"


// Console variables that we need to access from this module
cVar_t		*vid_xpos;			// X coordinate of window position
cVar_t		*vid_ypos;			// Y coordinate of window position
cVar_t		*vid_fullscreen;

// Global variables used internally by this module
//vidDef_t	viddef;				// global video state; used by other modules

struct {int width; int height; } viddef;

/*qBool	vid_ref_modified;*/
/*qBool	vid_ref_active;*/

#define VID_NUM_MODES ( sizeof( vid_modes ) / sizeof( vid_modes[0] ) )

extern void	VID_MenuShutdown (void);

static qBool    vid_queueRestart;
static qBool    vid_isActive;

/*
==========================================================================

DLL GLUE

==========================================================================
*/

void VID_Restart (void)
{
	vid_queueRestart = qTrue;
}

/*
============
VID_Restart_f

Console command to re-start the video mode and refresh DLL. We do this
simply by setting the modified flag for the vid_ref variable, which will
cause the entire video mode and refresh DLL to be reset on the next frame.
============
*/
void VID_Restart_f (void)
{
  VID_Restart();
}

/*
** VID_GetModeInfo
*/
typedef struct vidmode_s
{
	const char *description;
	int         width, height;
	int         mode;
} vidmode_t;

vidmode_t vid_modes[] =
{
	{ "Mode 0: 320x240",   320, 240,   0 },
	{ "Mode 1: 400x300",   400, 300,   1 },
	{ "Mode 2: 512x384",   512, 384,   2 },
	{ "Mode 3: 640x480",   640, 480,   3 },
	{ "Mode 4: 800x600",   800, 600,   4 },
	{ "Mode 5: 960x720",   960, 720,   5 },
	{ "Mode 6: 1024x768",  1024, 768,  6 },
	{ "Mode 7: 1152x864",  1152, 864,  7 },
	{ "Mode 8: 1280x1024",  1280, 1024, 8 },
	{ "Mode 9: 1600x1200", 1600, 1200, 9 },
	{ "Mode 10: 2048x1536", 2048, 1536, 10 }
};

qBool VID_GetModeInfo(int *width, int *height, int mode )
{
	if ( mode < 0 || mode >= VID_NUM_MODES )
		return qFalse;

	*width  = vid_modes[mode].width;
	*height = vid_modes[mode].height;

	return qTrue;
}

/*
** VID_NewWindow
*/
void VID_NewWindow(int width, int height)
{
	viddef.width  = width;
	viddef.height = height;
}

/*
============
VID_CheckChanges

This function gets called once just before drawing each frame, and it's sole purpose in life
is to check to see if any of the video mode parameters have changed, and if they have to 
update the rendering DLL and/or video mode to match.
============
*/
void    VID_CheckChanges (glConfig_t *outConfig)
{
	int errNum;

        while (vid_queueRestart) {
                qBool cgWasActive = cls.mapLoaded;

                CL_ShutdownMedia ();

                // Refresh has changed
                vid_queueRestart = qFalse;
                cls.refreshPrepped = qFalse;
                cls.disableScreen = qTrue;

                // Kill if already active
		if (vid_isActive)
		  R_Shutdown (), vid_isActive = qFalse;

                // Initialize renderer
                errNum = R_Init (NULL, NULL);

                // Refresh init failed!
                if (errNum != R_INIT_SUCCESS) {
                        R_Shutdown ();
                        vid_isActive = qFalse;

                        switch (errNum) {
                        case R_INIT_QGL_FAIL:
                                Com_Error (ERR_FATAL, "Couldn't initialize OpenGL!\n" "QGL library failure!");
                                break;

                        case R_INIT_OS_FAIL:
                                Com_Error (ERR_FATAL, "Couldn't initialize OpenGL!\n" "Incorrect operating system!");
                                break;

                        case R_INIT_MODE_FAIL:
                                Com_Error (ERR_FATAL, "Couldn't initialize OpenGL!\n" "Couldn't set video mode!");
                                break;
                        }
                }

                R_GetGLConfig (outConfig);

                Snd_Init ();
                CL_InitMedia ();

                IN_Restart_f ();

                cls.disableScreen = qFalse;

                Con_Close ();

                // This is to stop cgame from initializing on first load
                // and so it will load after a vid_restart while connected somewhere
                if (cgWasActive) {
                        CL_CGModule_LoadMap ();
                        Key_SetKeyDest (KD_GAME);
                }
                else {
                        CL_CGModule_MainMenu ();
                }

                vid_isActive = qTrue;
        }
}

/*
============
VID_Init
============
*/
void    VID_Init (glConfig_t *outConfig)
{
	vid_xpos = Cvar_Register ("vid_xpos", "3", CVAR_ARCHIVE);
	vid_ypos = Cvar_Register ("vid_ypos", "22", CVAR_ARCHIVE);
	vid_fullscreen = Cvar_Register ("vid_fullscreen", "0", CVAR_ARCHIVE);

	/* Add some console commands that we want to handle */
	Cmd_AddCommand ("vid_restart", VID_Restart_f,		"Restarts refresh and media");

	/* Start the graphics mode and load refresh DLL */
	vid_isActive = qFalse;
	vid_fullscreen->modified = qTrue;
	vid_queueRestart = qTrue;
	VID_CheckChanges (outConfig);
}

/*
============
VID_Shutdown
============
*/
void VID_Shutdown (void)
{
	if(vid_isActive) {
		R_Shutdown ();
		vid_isActive = qFalse;
	}

	Cmd_RemoveCommand ("vid_restart", NULL);
}


