/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

//
// r_local.h
// Refresh only header file
//

#define SHADOW_VOLUMES	2
#define SHADOW_ALPHA	0.5f

#ifdef _WIN32
# include <windows.h>
#endif

#include <GL/gl.h>
#include <math.h>

#include "refresh.h"
#include "qgl.h"

#include "r_image.h"
#include "r_program.h"
#include "r_shader.h"
#include "r_entity.h"
#include "r_backend.h"
#include "r_model.h"
#include "r_glext.h"
#include "r_glstate.h"

/*
=============================================================================

	REFRESH

=============================================================================
*/

enum {
	REND_CLASS_DEFAULT,
	REND_CLASS_MCD,

	REND_CLASS_3DLABS_GLINT_MX,
	REND_CLASS_3DLABS_PERMEDIA,
	REND_CLASS_3DLABS_REALIZM,
	REND_CLASS_ATI,
	REND_CLASS_ATI_RADEON,
	REND_CLASS_INTEL,
	REND_CLASS_NVIDIA,
	REND_CLASS_NVIDIA_GEFORCE,
	REND_CLASS_PMX,
	REND_CLASS_POWERVR_PCX1,
	REND_CLASS_POWERVR_PCX2,
	REND_CLASS_RENDITION,
	REND_CLASS_S3,
	REND_CLASS_SGI,
	REND_CLASS_SIS,
	REND_CLASS_VOODOO
};

typedef struct glMedia_s {
	shader_t		*charShader;

	shader_t		*worldLavaCaustics;
	shader_t		*worldSlimeCaustics;
	shader_t		*worldWaterCaustics;
} glMedia_t;

typedef struct glStatic_s {
	short			renderClass;

	const char		*rendererString;
	const char		*vendorString;
	const char		*versionString;
	const char		*extensionString;

#ifdef _WIN32
	qBool			allowCDS;
#endif // _WIN32

	// Resolution
	int				safeMode;

	// Hardware gamma
	qBool			rampDownloaded;
	uShort			originalRamp[768];
	uShort			gammaRamp[768];

	// Texture
	int				rgbFormat;
	int				rgbaFormat;

	int				rgbFormatCompressed;
	int				rgbaFormatCompressed;

	// PFD Stuff
	qBool			useStencil;

	byte			cAlphaBits;
	byte			cColorBits;
	byte			cDepthBits;
	byte			cStencilBits;
} glStatic_t;

extern glConfig_t	glConfig;
extern glMedia_t	glMedia;
extern glStatic_t	glStatic;

/*
=============================================================================

	PERFORMANCE COUNTERS

=============================================================================
*/

typedef struct refStats_s {
	// Totals
	uInt			numTris;
	uInt			numVerts;
	uInt			numElements;

	uInt			meshCount;
	uInt			meshPasses;

	// Alias Models
	uInt			aliasElements;
	uInt			aliasPolys;

	// Culling
	uInt			cullPass;
	uInt			cullFail;
	uInt			cullTotal;

	// Image
	uInt			texInUse;
	uInt			texBinds;
	uInt			texEnvChanges;
	uInt			texUnitChanges;

	// World model
	uInt			worldElements;
	uInt			worldPolys;
} refStats_t;

extern refStats_t	r_refStats;

/*
=============================================================================

	SCENE

=============================================================================
*/

typedef struct refScene_s {
	uInt			frameCount;
	uInt			registerFrame;

	// View
	cBspPlane_t		viewFrustum[4];

	mat4x4_t		modelViewMatrix;
	mat4x4_t		projectionMatrix;
	mat4x4_t		worldViewMatrix;

	qBool			mirrorView;
	qBool			portalView;
	vec3_t			portalOrigin;
	cBspPlane_t		clipPlane;

	float			zFar;

	// World model
	model_t			*worldModel;
	entity_t		worldEntity;

	uInt			visFrameCount;	// Bumped when going to a new PVS
	int				viewCluster;
	int				oldViewCluster;

	// Items
	int				numEntities;
	entity_t		entityList[MAX_ENTITIES];

	int				numPolys;
	poly_t			polyList[MAX_POLYS];

	int				numDLights;
	dLight_t		dLightList[MAX_DLIGHTS];

	lightStyle_t	lightStyles[MAX_CS_LIGHTSTYLES];
} refScene_t;

extern refDef_t		r_refDef;
extern refScene_t	r_refScene;

void	R_RenderToList (refDef_t *rd, meshList_t *list);

void	GL_CheckForError (char *where);

/*
=============================================================================

	CVARS

=============================================================================
*/

extern cVar_t	*e_test_0;
extern cVar_t	*e_test_1;

extern cVar_t	*con_font;

extern cVar_t	*intensity;

extern cVar_t	*gl_3dlabs_broken;
extern cVar_t	*gl_bitdepth;
extern cVar_t	*gl_clear;
extern cVar_t	*gl_cull;
extern cVar_t	*gl_drawbuffer;
extern cVar_t	*gl_driver;
extern cVar_t	*gl_dynamic;
extern cVar_t	*gl_errorcheck;

extern cVar_t	*r_ext_maxAnisotropy;

extern cVar_t	*gl_finish;
extern cVar_t	*gl_flashblend;
extern cVar_t	*gl_jpgquality;
extern cVar_t	*gl_lightmap;
extern cVar_t	*gl_lockpvs;
extern cVar_t	*gl_log;
extern cVar_t	*gl_maxTexSize;
extern cVar_t	*gl_mode;
extern cVar_t	*gl_modulate;

extern cVar_t	*gl_polyblend;
extern cVar_t	*gl_screenshot;
extern cVar_t	*gl_shadows;
extern cVar_t	*gl_shownormals;
extern cVar_t	*gl_showtris;
extern cVar_t	*gl_stencilbuffer;
extern cVar_t	*gl_texturebits;
extern cVar_t	*gl_texturemode;

extern cVar_t	*qgl_debug;

extern cVar_t	*r_bumpScale;
extern cVar_t	*r_caustics;
extern cVar_t	*r_detailTextures;
extern cVar_t	*r_displayFreq;
extern cVar_t	*r_drawEntities;
extern cVar_t	*r_drawworld;
extern cVar_t	*r_facePlaneCull;
extern cVar_t	*r_flares;
extern cVar_t	*r_flareFade;
extern cVar_t	*r_flareSize;
extern cVar_t	*r_fontscale;
extern cVar_t	*r_fullbright;
extern cVar_t	*r_hwGamma;
extern cVar_t	*r_lerpmodels;
extern cVar_t	*r_lightlevel;	// FIXME: This is a HACK to get the client's light level
extern cVar_t	*r_lmMaxBlockSize;
extern cVar_t	*r_lmModulate;
extern cVar_t	*r_lmPacking;
extern cVar_t	*r_noCull;
extern cVar_t	*r_noRefresh;
extern cVar_t	*r_noVis;
extern cVar_t	*r_patchDivLevel;
extern cVar_t	*r_showCull;
extern cVar_t	*r_speeds;
extern cVar_t	*r_sphereCull;
extern cVar_t	*r_swapInterval;
extern cVar_t	*r_vertexLighting;

extern cVar_t	*r_alphabits;
extern cVar_t	*r_colorbits;
extern cVar_t	*r_depthbits;
extern cVar_t	*r_stencilbits;
extern cVar_t	*cl_stereo;
extern cVar_t	*gl_allow_software;
extern cVar_t	*gl_stencilbuffer;

extern cVar_t	*vid_gammapics;
extern cVar_t	*vid_gamma;
extern cVar_t	*vid_width;
extern cVar_t	*vid_height;

extern cVar_t	*intensity;

extern cVar_t	*gl_jpgquality;
extern cVar_t	*gl_nobind;
extern cVar_t	*gl_picmip;
extern cVar_t	*gl_screenshot;
extern cVar_t	*gl_texturebits;
extern cVar_t	*gl_texturemode;

/*
=============================================================================

	SUPPORTING FUNCTIONS

=============================================================================
*/

//
// console.c
//

void	Con_CheckResize (void);

//
// r_cin.c
//

void	R_CinematicPalette (const byte *palette);
void	R_DrawCinematic (int x, int y, int w, int h, int cols, int rows, byte *data);

//
// r_cull.c
//

void	R_SetupFrustum (void);

qBool	R_CullBox (vec3_t mins, vec3_t maxs, int clipFlags);
qBool	R_CullSphere (const vec3_t origin, const float radius, int clipFlags);

qBool	R_CullVisNode (struct mBspNode_s *node);

//
// r_draw.c
//

void	R_CheckFont (void);

void	R_DrawPic (shader_t *shader, float x, float y, int w, int h, float s1, float t1, float s2, float t2, vec4_t color);
void	R_DrawChar (shader_t *charsShader, float x, float y, int flags, float scale, int num, vec4_t color);
int		R_DrawString (shader_t *charsShader, float x, float y, int flags, float scale, char *string, vec4_t color);
int		R_DrawStringLen (shader_t *charsShader, float x, float y, int flags, float scale, char *string, int len, vec4_t color);
void	R_DrawFill (float x, float y, int w, int h, vec4_t color);

//
// r_init.c
//

qBool	R_GetInfoForMode (int mode, int *width, int *height);

//
// r_light.c
//

void	R_Q2BSP_MarkLights (qBool visCull, dLight_t *light, int bit, mBspNode_t *node);

void	R_AddDLightsToList (void);

void	R_Q2BSP_PushDLights (void);

void	LM_Q2BSP_UpdateLightmap (mBspSurface_t *surf);
void	LM_Q2BSP_BeginBuildingLightmaps (void);
void	LM_Q2BSP_CreateSurfaceLightmap (mBspSurface_t *surf);
void	LM_Q2BSP_EndBuildingLightmaps (void);

void	R_LightBounds (const vec3_t origin, float intensity, vec3_t mins, vec3_t maxs);
void	R_LightPoint (vec3_t point, vec3_t light);
void	R_SetLightLevel (void);

qBool	R_ShadowForEntity (entity_t *ent, vec3_t shadowSpot);
void	R_LightForEntity (entity_t *ent, int numVerts, byte *bArray);

//
// r_lightq3.c
//

void	R_Q3BSP_MarkLights (void);
void	R_Q3BSP_MarkLightsBModel (entity_t *ent, vec3_t mins, vec3_t maxs);
void	R_Q3BSP_AddDynamicLights (uInt dLightBits, entity_t *ent);

void	R_Q3BSP_LightForEntity (entity_t *ent, int numVerts, byte *bArray);

void	R_Q3BSP_BuildLightmaps (int numLightmaps, int w, int h, const byte *data, mlightmapRect_t *rects);
void	R_Q3BSP_TouchLightmaps (void);

//
// r_math.c
//

byte	R_FloatToByte (float x);

float	R_ColorNormalize (const float *in, float *out);

void	Matrix4_Copy2D (const mat4x4_t m1, mat4x4_t m2);
void	Matrix4_Multiply2D (const mat4x4_t m1, const mat4x4_t m2, mat4x4_t out);
void	Matrix4_Scale2D (mat4x4_t m, float x, float y);
void	Matrix4_Stretch2D (mat4x4_t m, float s, float t);
void	Matrix4_Translate2D (mat4x4_t m, float x, float y);

void	Matrix4_Multiply_Vector (const mat4x4_t m, const vec4_t v, vec4_t out);

//
// r_poly.c
//

void	R_AddPolysToList (void);
void	R_PushPoly (meshBuffer_t *mb);
qBool	R_PolyOverflow (meshBuffer_t *mb);
void	R_DrawPoly (meshBuffer_t *mb);

//
// r_shadow.c
//

#ifdef SHADOW_VOLUMES
void	R_DrawShadowVolumes (mesh_t *mesh, entity_t *ent, vec3_t mins, vec3_t maxs, float radius);
void	R_ShadowBlend (void);
#endif

void	R_SimpleShadow (entity_t *ent, vec3_t shadowSpot);

//
// r_sky.c
//

#define SKY_MAXCLIPVERTS	64
#define SKY_BOXSIZE			8192

void	R_ClipSkySurface (mBspSurface_t *surf);
void	R_AddSkyToList (void);

void	R_ClearSky (void);
void	R_DrawSky (meshBuffer_t *mb);

void	R_SetSky (char *name, float rotate, vec3_t axis);

void	R_SkyInit (void);
void	R_SkyShutdown (void);

//
// r_world.c
//

void	R_AddWorldToList (void);
void	R_AddQ2BrushModel (entity_t *ent);

void	R_WorldInit (void);
void	R_WorldShutdown (void);

//
// r_worldq3.c
//

void	R_AddQ3WorldToList (void);
void	R_AddQ3BrushModel (entity_t *ent);

/*
=============================================================================

	IMPLEMENTATION SPECIFIC

=============================================================================
*/

extern cVar_t	*vid_fullscreen;
extern cVar_t	*vid_xpos;
extern cVar_t	*vid_ypos;

//
// glimp_imp.c
//

void	GLimp_BeginFrame (void);
void	GLimp_EndFrame (void);

void	GLimp_AppActivate (qBool active);
qBool	GLimp_GetGammaRamp (uShort *ramp);
void	GLimp_SetGammaRamp (uShort *ramp);

void	GLimp_Shutdown (void);
qBool	GLimp_Init (void *hInstance, void *hWnd);
qBool	GLimp_AttemptMode (int mode);
